# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: copyfrom [-rpC] -c CLUSTER [-u USER] FILE LOCAL_DIR

Copies FILE from each node of the specified CLUSTER to LOCAL_DIR.
If CLUSTER is omitted, then the default cluster is used. CLUSTER
may also name a single node.

Each FILE must be an absolute path, (i.e. must begin with '/').

copyfrom will create a directory under LOCAL_DIR for each node in
the cluster.  The name of the directory will be the same as the
node's name.
    
The flags are a subset of the options supported by scp:
    -r: recursive copy
    -p: preserve modification times, access times, and modes.
    -C: enable compression.
"""

import os
import pickle
import popen2
import threading

import osh.oshargs
import osh.oshobjects

def copyfrom(*args):
    if len(args) > 0:
        cluster = args[0]
        args = (osh.oshargs.Option('-c', cluster),) + args[1:]
    return _CopyFrom().process_args(*args)

# scp options, also used by copyto

def scp(options):
    if options[0] != '-':
        options = '-' + options
    return osh.oshargs.Option(options)

class _CopyFrom(osh.oshobjects.RemoteOp):

    # state

    _local_dir = None
    _file = None
    _scp_options = None


    # object interface
    
    def __init__(self):
        osh.oshobjects.RemoteOp.__init__(self, 'rpCc:u:', (2, 2))


    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.oshobjects.RemoteOp.setup(self)
        args = self.args()
        if args.has_next():
            self._file = args.next_string()
        if args.has_next():
            self._local_dir = args.next_string()
        if not self._file or not self._local_dir or args.has_next():
            self.usage()
        if not self._file.startswith('/'):
            self.usage()

    # RemoteOp interface

    def action(self, host):
        target_dir = self._local_dir + '/' + host.name()
        try:
            os.makedirs(target_dir)
        except OSError:
            pass
        osh.oshobjects.copydown(self.user(),
                                host.address(),
                                self._file,
                                target_dir,
                                self.scp_options())
