# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: out [-t] [-a FILENAME] [-f FILENAME] [-c | FORMAT]

Input objects are passed on to the output stream. Objects are
also printed to stdout, or if FILENAME is specified, to the named
file. The FILENAME may be specified using the -a flag or the -f
flag, but not both. -a will append the file, while -f will overwrite.

The output string is generated by applying str() to the object, unless
FORMAT is specified, in which case the object is formatted using the %
operator. -c is another formatting option, specifying that output
should be generated in CSV format (comma-separated values).

-t indicates that out is 'terminal' -- no objects are sent to the output
stream."""

import sys

import osh.oshargs
import osh.oshobjects

def out(*args):
    return _Out().process_args(*args)

def append(filename):
    return osh.oshargs.Option('-a', filename)

def write(filename):
    return osh.oshargs.Option('-f', filename)

def csv():
    return osh.oshargs.Option('-c')

def terminal():
    return osh.oshargs.Option('-t')

class _Out(osh.oshobjects.Op):

    # state

    _format = None
    _csv = False
    _output = None
    _terminal = None


    # object interface

    def __init__(self):
        osh.oshobjects.Op.__init__(self, 'a:cf:t', (0, 1))


    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        append_filename = args.string_arg('-a')
        replace_filename = args.string_arg('-f')
        filename = None
        mode = None
        if append_filename and replace_filename:
            self.usage()
        elif append_filename:
            mode = 'a'
            filename = append_filename
        elif replace_filename:
            mode = 'w'
            filename = replace_filename
        self._csv = args.flag('-c')
        self._terminal = args.flag('-t')
        if args.has_next():
            self._format = args.next_string()
            if args.has_next():
                self.usage()
        if self._csv and self._format:
            self.usage()
        if filename:
            self._output = open(filename, mode)
        else:
            self._output = sys.stdout
    
    def receive(self, object):
        if self._format:
            try:
                formatted_object = self._format % object
            except Exception, e:
                # If there is one %s in the format, and the object is longer,
                # then convert it to a string
                if self._format.count('%') == 1 and self._format.count('%s') == 1:
                    formatted_object = self._format % str(object)
                else:
                    raise e
        elif self._csv:
            if type(object) in (list, tuple):
                formatted_object = ', '.join([_quote_unless_number(x) for x in object])
            else:
                formatted_object = str(object)
        else:
            if type(object) in (list, tuple):
                formatted_object = '(' + ', '.join([_quote_unless_number(x) for x in object]) + ')'
            else:
                formatted_object = str(object)
        print >> self._output, formatted_object
        self._output.flush()
        if not self._terminal:
            self.send(object)

    def receive_complete(self):
        if self._output != sys.stdout:
            self._output.close()

def _quote_unless_number(x):
    if type(x) in (int, long, float):
        return str(x)
    else:
        return '"%s"' % x
