# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: sh 'COMMAND'
Escape to a native shell and execute COMMAND. stdout and sterr from
COMMAND are sent to streams o and e respectively, with each line of
output giving rise to a string. Occurrences of %s in COMMAND will be
replaced by piped in values.
"""

import popen2
import threading

import osh.oshobjects

def sh(*args):
    return _Sh().process_args(*args)

class _Sh(osh.oshobjects.Generator):

    # state

    _command = None


    # object interface
    
    def __init__(self):
        osh.oshobjects.Generator.__init__(self, '', (1, 1))


    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        if args.has_next():
            self._command = args.next_string()
            if args.has_next():
                self.usage()
        else:
            self.usage()

    def receive(self, object):
        boundCommand = self._bind(object)
        self._execute_command(boundCommand)

    def receive_complete(self):
        self.send_complete()

    # remote compile-time interface

    def setCommand(self, command):
        self._command = command
        return self


    # Generator interface
    
    def execute(self):
        self._execute_command(self._command)
        self.send_complete()




    # For use by this class

    def _bind(self, input):
        command = self._command
        if type(input) != tuple:
            input = (input,)
        for value in input:
            command = command.replace('%s', str(value), 1)
        return command

    def _execute_command(self, command):
        process = popen2.Popen3(command, True)
        stdout_consumer = _StreamConsumer(process.fromchild,
                                          lambda s: self.send(s))
        stderr_consumer = _StreamConsumer(process.childerr,
                                          lambda s: self.send_error(s))
        stdout_consumer.start()
        stderr_consumer.start()
        process.wait()
        stdout_consumer.join()
        stderr_consumer.join()

class _StreamConsumer(threading.Thread):
    _process = None
    _stream = None
    _output_handler = None

    def __init__(self, stream, output_handler):
        threading.Thread.__init__(self)
        self._stream = stream
        self._output_handler = output_handler

    def run(self):
        eof = False
        while not eof:
            line = self._stream.readline()
            if line:
                # get rid of newline
                if line.endswith('\n'):
                    self._output_handler(line[:-1])
                else:
                    self._output_handler(line)
            else:
                eof = True
        self._stream.close()

