# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

from __future__ import generators
import os
import signal
import sys
import traceback
import types

import osherror
import oshprocess

# osh configuration

oshConfiguration = {}

def wrap_if_necessary(object):
    if type(object) is not types.TupleType and type(object) is not types.ListType:
        object = (object,)
    return object

class Config(object):
    """key/value pairs used to configure osh. key is a string of
    dot-separated tokens with these properties:
    - First token is 'osh'.
    - All tokens are legal python identifiers.
    The value may be any python object.
    """
    
    _tokens = None

    def __init__(self):
        # Can't assign to self._tokens because __setattr__ would be invoked.
        object.__setattr__(self, '_tokens', [])

    def __getattr__(self, token):
        self._tokens.append(token)
        return self

    def __setattr__(self, token, value):
        self._tokens.append(token)
        oshConfiguration[self.key()] = value

    def __getitem__(self, key):
        self._tokens.append(key)
        return self

    def __setitem__(self, key, value):
        self._tokens.append(key)
        oshConfiguration[self.key()] = value

    def key(self):
        return '.'.join(self._tokens)

class _ConfigFactory:

    def __getattr__(self, token):
        return Config().__getattr__(token)

    def __setattr__(self, token, value):
        return Config().__setattr__(token, value)

    def __getitem__(self, key):
        return Config().__getitem__(key)

    def __setitem__(self, key, value):
        return Config().__setattr__(key, value)

osh = _ConfigFactory()

def _default_profile(command):
    """The default profile name for an osh command. E.g., if osh.remote='foobar',
    then 'osh @ [ ... ]' is equivalent to 'osh @foobar [...]'.
    """
    return oshConfiguration[command]

def config_value(*args):
    """Returns the configuration value specified by args. E.g., if
    osh.a.b.c='xyz' then config_value('a', 'b', 'c') returns 'xyz'.
    """
    if len(args) == 1:
        configKey = args[0]
    elif len(args) == 3:
        command = args[0]
        profile = args[1]
        key = args[2]
        if profile is None:
            profile = _default_profile(command)
        configKey = '%s.%s.%s' % (command, profile, key)
    else:
        sys.stderr.write('Illegal arguments to oshconfig.get: %s' %
                         str(args))
    return oshConfiguration.get(configKey, None)

# Handle ctrl-c

def _ctrl_c_handler(signal, frame):
    oshprocess.kill_all_processes()
    sys.stdout.close()
    sys.stderr.close()
    os.kill(os.getpid(), 9) # Kill self
    sys.exit(0) # Unreachable if suicide worked
    
signal.signal(signal.SIGINT, _ctrl_c_handler)

# Read user's configuration

def findconfig():
    # Check local dir first (.oshrc)
    # Check home dir second (.oshrc)
    # Check etc dir third (oshrc)
    possible = [ '.oshrc' ]
    if os.getenv('HOME'):
        possible.append(os.getenv('HOME') + '/.oshrc')
    possible.append('/etc/oshrc')
    oshrcPath = None
    possible = [ '.oshrc' ]
    if os.getenv('HOME'):
        possible.append(os.getenv('HOME') + '/.oshrc')
    possible.append('/etc/oshrc')

    for f in possible:
        if os.path.isfile(f):
            oshrcPath = f
            break

    if not oshrcPath:
        oshrcPath = os.getenv('HOME') + '/.oshrc'
        oshrcFile = open(oshrcPath, 'w')
        print >>oshrcFile, 'from osh.oshconfig import *\n'
        oshrcFile.close()

    return oshrcPath

# osh environment

OSH_HOME = '/usr/share/osh'
OSH_VERSION = '0.8.1'
OSH_SUBVERSION = '2(cm)'

# global constants and variables

STREAM_OUTPUT = 'o'
STREAM_ERROR = 'e'
verbosity = 0

def exception_handler(receiver, command, input_object, exception):
    if receiver:
        receiver.receive(STREAM_ERROR, wrap_if_necessary(osherror.OshError(str(command), input_object, exception)))
    else:
        raise Exception('receiver undefined, command: %s, input_object: %s, error: %s' %
                        (command, input_object, exception))
    if verbosity >= 2:
        traceback.print_exc(file = sys.stderr)


# Import types needed by builtin osh commands

from oshtypes.process import *

# builtins available to all OshFunctions

generators = {}

def _number_generator(initialValue):
    n = initialValue
    while True:
        yield n
        n = n + 1

def n(functionId, initialValue):
    generator = generators.get(functionId, None)
    if generator is None:
        generator = _number_generator(initialValue)
        generators[functionId] = generator
    return generator.next()

def ifelse(predicate, if_true, if_false):
    """Returns if_true if predicate is true, if_false otherwise.
    """
    if predicate:
        return if_true
    else:
        return if_false

def oshversion():
    return OSH_VERSION

execfile(findconfig())

# osh package path (set this after reading config file)

oshpath = config_value('path')
if not oshpath:
    oshpath = []

