# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Objects comprising an osh command's parse tree. Also
includes some classes used at command runtime.
"""

from __future__ import generators
import getopt
import new
import os
import popen2
import threading
import types
import traceback

from oshconfig import *
import oshargs

USAGE_API = 'api'
USAGE_CLI = 'cli'
# Assume API. If oshparser is used then usage will be changed.
osh_usage = USAGE_API

# scp options supported by copyfrom and copyto
SCP_OPTIONS = 'rpC'

# 2.2 threads sometimes hang without this. The 2.2 default is 10, 2.3: 100
sys.setcheckinterval(100)

verbosity = None
default_db_profile = None

_id_counter = 0
_id_counter_lock = threading.Condition()

class ErrorHandlerException(Exception):

    def __init__(self, reason):
        Exception.__init__(self)
        self.reason = reason

def _call_exception_handler(exception_handler, receiver, command, input_object, exception):
    try:
        exception_handler(receiver, command, input_object, exception)
    except ErrorHandlerException, e:
        raise
    except Exception, e:
        raise ErrorHandlerException(e)

def _next_id():
    global _id_counter
    global _id_counter_lock
    _id_counter_lock.acquire()
    _id_counter += 1
    next_id = _id_counter
    _id_counter_lock.release()
    return next_id

#--------------------------------------------------------------------------------

class OshObject:
    """Base class for all objects in a parse tree representing an osh
    command.
    """

    # state

    _id = None


    # object interface

    def __init__(self):
        self._id = _next_id()


    # OshObject interface

    def id(self):
        """Unique identifier within a single command.
        """
        return self._id

    def dump(self):
        """Renders this object and subtree as a string for debugging purposes.
        """
        assert False

#--------------------------------------------------------------------------------

class BaseOp(OshObject):
    """Base class for all osh commands, and for pipelines, (sequences of
    commands). Methods of this class implement the command execution and
    inter-command communication. The send* commands are used by subclasses to
    send output to downstream commands. The receive* commands are implemented
    by subclasses to receive and process input from upstream commands.
    """

    # state

    _label = None
    _parentOpset = None


    # object interface
    
    def __init__(self):
        OshObject.__init__(self)
        self._label = STREAM_OUTPUT

    def __repr__(self):
        assert False


    # BaseOp compile-time interface
    
    def set_opset(self, opset):
        self._parentOpset = opset


    # BaseOp runtime interface

    def generator(self):
        """Returns True if this class is a generator, False otherwise.
        (A generator is an osh command that does not require any input.)
        """
        return False

    def usage(self):
        """Print usage message and exit.
        """
        sys.stderr.write(self.doc())
        sys.stderr.write('\n')
        sys.exit(-1)

    def doc():
        """Print command usage information.
        """
        assert False
    
    def setup(self):
        """Setup is executed just prior to command execution.
        """
        pass

    def execute(self):
        """Execute the command.
        """
        assert False

    def send(self, object, label = STREAM_OUTPUT):
        """Called by a command class to send an object of command output to
        the next command, via the stream specified by the label, ('o' by default).
        """
        self._parentOpset.send(wrap_if_necessary(object), label)

    def send_error(self, object):
        """Called by a command class to send an object describing an error to
        the 'e' (error) stream.
        """
        self._parentOpset.send(wrap_if_necessary(object), STREAM_ERROR)

    def send_complete(self):
        """Called by a command class to indicate that there will
        be no more output from the command.
        """
        self._parentOpset.send_complete()

    def receive(self, object):
        """Implemented by a command class to process an input object.
        """
        pass

    def receive_complete(self):
        """Implemented by a command class to do any cleanup required
        after all input has been received.
        """
        pass


    # For use by this package

    def parent_opset(self):
        return self._parentOpset

    def label(self):
        return self._label

    def set_label(self, label):
        self._label = label


#--------------------------------------------------------------------------------

class Op(BaseOp):
    """Base class for all osh commands, (excluding pipelines).
    """

    # state
    
    _args = None


    # object interface
    
    def __init__(self, flag_spec, anon_range):
        BaseOp.__init__(self)
        if osh_usage == USAGE_API:
            self._args = oshargs.OshArgsAPI(self, flag_spec, anon_range)
        elif osh_usage == USAGE_CLI:
            self._args = oshargs.OshArgsCLI(self, flag_spec, anon_range)
        else:
            assert False

    def __repr__(self):
        return '%s#%s%s' % (self._command_name(), self.id(), self._args)


    # OshObject interface
    
    def dump(self):
        return str(self)


    # BaseOp interface

    def execute(self):
        assert False


    # Op compile-time interface
    
    def add_arg(self, arg):
        self._args.add_token(arg)
        return self

    def args_done(self):
        self._args.args_done()
        return self


    # Op interface - for use by subclasses

    def process_args(self, *args):
        if osh_usage == USAGE_API:
            for arg in args:
                self._args.add_arg(arg)
            self._args.args_done()
            # print 'process_args: %s' % self._args
        # else: args will be supplied from parser using add_arg and
        # processed by setup method.
        return self

    def args(self):
        return self._args
    
    
    # for use by this class

    def _command_name(self):
        return str(self.__class__).split('.')[-1]

#--------------------------------------------------------------------------------

class Pipeline(BaseOp):

    # state

    _first_opset = None
    

    # object interface

    def __init__(self, opset = None):
        BaseOp.__init__(self)
        if opset:
            self.append_opset(opset)

    def __repr__(self):
        return 'pipeline#%s' % self.id()


    # OshObject interface

    def dump(self):
        return '%s(%s)' % (self, self._first_opset.dump())


    # BaseOp runtime interface

    def generator(self):
        return self._first_opset.generator()

    def execute(self):
        self._first_opset.execute()

    def receive(self, object):
        self._first_opset.receive(STREAM_OUTPUT, object)

    def receive_complete(self):
        self._first_opset.receive_complete()
    

    # Pipeline compile-time interface

    def append_opset(self, opset):
        if self._first_opset:
            self._first_opset.connect(opset)
        else:
            self._first_opset = opset
        opset.set_parent_pipeline(self)
        return self

    def prepend_opset(self, opset):
        if self._first_opset:
            self._first_opset = opset.connect(self._first_opset)
        else:
            self._first_opset = opset
        opset.set_parent_pipeline(self)
        return self


    # Pipeline runtime interface

    def setup(self):
        self._first_opset.setup()

    def first_opset(self, opset):
        return self._first_opset == opset


#--------------------------------------------------------------------------------

class Opset(OshObject):

    # state

    _ops = None # map: label -> op
    _next_opset = None
    _parent_pipeline = None
    _receiver = None


    # object interface
    
    def __init__(self, op = None):
        OshObject.__init__(self)
        self._ops = {}
        if op:
            self.add_op(op)

    def __repr__(self):
        return 'opset#%s' % self.id()


    # OshObject interface

    def dump(self):
        tokens = [str(self), '(']
        ops = {}
        first = True
        for label, op in self._ops.iteritems():
            if first:
                first = False
            else:
                tokens.append(', ')
            tokens.append(label)
            tokens.append(': ')
            tokens.append(op.dump())
        tokens.append(')')
        if self._next_opset:
            tokens.append(' ^ ')
            tokens.append(self._next_opset.dump())
        return ''.join(tokens)


    # Opset compile-time interface
    
    def add_op(self, op):
        self._ops[op.label()] = op
        op.set_opset(self)
        return self

    def connect(self, new_opset):
        last_opset  = self
        while last_opset._next_opset:
            last_opset = last_opset._next_opset
        last_opset._next_opset = new_opset
        return self

    def set_parent_pipeline(self, pipeline):
        self._parent_pipeline = pipeline


    # op runtime interface

    def setup(self):
        self._receiver = self._opset_receiver()
        # Add an error handler if there is none, unless this opset is
        # the first of a pipeline. (In which case, there is no previous
        # opset to cause an error.)
        if (not self._parent_pipeline.first_opset(self) and
            not self._ops.get(STREAM_ERROR)):
            import command.out
            if osh_usage == USAGE_CLI:
                out_op = command.out.out()
                out_op.add_arg('-t')
                out_op.add_arg('ERROR: %s')
                out_op.args_done()
            elif osh_usage == USAGE_API:
                # out_op = command.out.out(oshargs.Option('-t'), 'ERROR: %s')
                import command.blowup
                out_op = command.blowup.blowup()
            else:
                assert False
            out_op.set_label(STREAM_ERROR)
            self.add_op(out_op)
        for op in self._ops.values():
            op.setup()
        if self._next_opset:
            self._next_opset.setup()

    def execute(self):
        assert len(self._ops) == 1, (
            'Attempt to execute opset ' +
            'containing != 1 generator: %s' % (self._ops))
        generator = self._ops.values()[0]
        try:
            generator.execute()
        except ErrorHandlerException:
            raise
        except Exception, e:
            # generator.execute threw an exception. The error
            # handler is downstream from generator.
            _call_exception_handler(
                exception_handler,
                self._receiver,
                generator,
                None,
                e)

    def send(self, object, label = STREAM_OUTPUT):
        if self._receiver:
            if label == STREAM_ERROR and osh_usage == USAGE_API:
                # The API doesn't pass on exceptions, it raises them.
                # object[0] is an exception (since label is e)
                # raise object[0]
                pass
            try:
                self._receiver.receive(label, object)
            except ErrorHandlerException:
                raise
            except Exception, e:
                # self._receiver.receive threw an exception. The error
                # handler is downstream from self._receiver.
                _call_exception_handler(
                    exception_handler,
                    self._receiver._receiver,
                    self._receiver.op(label),
                    object,
                    e)

    def last_opset(self):
        last = self
        while not (last._next_opset is None):
            last = last._next_opset
        return last

    def send_complete(self):
        if self._receiver:
            self._receiver.receive_complete()

    def receive(self, label, object):
        op = self._ops.get(label, None)
        if op:
            if verbosity >= 2:
                print '%s: %s -> %s' % (label, object, op)
            op.receive(object)

    def receive_complete(self):
        for op in self._ops.values():
            op.receive_complete()
        if self._receiver:
            self._receiver.receive_complete()

    def op(self, label):
        return self._ops.get(label, None)

    def ops(self):
        return self._ops
    
    def generator(self):
        if len(self._ops) > 1:
            generator = False
        else:
            generator = self._ops.values()[0].generator()
        return generator        


    # for use by this class

    def _opset_receiver(self):
        receiver = self._next_opset
        if (not receiver and
            self._parent_pipeline and
            self._parent_pipeline.parent_opset()):
            receiver = self._parent_pipeline.parent_opset()._opset_receiver()
        return receiver

    def generator(self):
        if len(self._ops) > 1:
            generator = False
        else:
            generator = self._ops.values()[0].generator()
        return generator        


#--------------------------------------------------------------------------------

class Generator(Op):
    """Base class of osh commands that are generators, e.g. gen, timer.
    """

    # object interface

    def __init__(self, flag_spec, anon_range):
        Op.__init__(self, flag_spec, anon_range)


    # BaseOp interface

    def generator(self):
        return True

    # Allows one generator to be run under the control of another
    # (e.g. timer)
    def receive(self, object):
        self.execute()


    # Generator interface

    def execute(self):
        assert False

#--------------------------------------------------------------------------------

class command:
    _pipeline = None

    def __init__(self, pipeline):
        self._pipeline = pipeline

    def dump(self):
        return self._pipeline.dump()

    def execute(self):
        # Prepare for execution.
        self._pipeline.setup()
        if verbosity >= 1:
            print self._pipeline.dump()
        # Execute.
        self._pipeline.execute()

    def pipeline(self):
        return self._pipeline

#--------------------------------------------------------------------------------


def _cluster_configuration(cluster):
    config_user = config_value('remote', cluster, 'user')
    config_hosts = config_value('remote', cluster, 'hosts')
    if type(config_hosts) is list or type(config_hosts) is tuple:
        hosts = [_Host(addr) for addr in config_hosts]
    elif type(config_hosts) is dict:
        hosts = []
        for name, host_spec in config_hosts.iteritems():
            host = None
            if type(host_spec) is str:
                host = _Host(name, host_spec)
            elif type(host_spec) is dict:
                addr = host_spec.get('host', None)
                db_profile = host_spec.get('db_profile', None)
                if addr is not None:
                    host = _Host(name, addr, db_profile)
            if host is None:
                raise Exception(('Error in ~/.oshrc: ' +
                                 'host specification in osh.remote.%s.hosts ' +
                                 'must be string, or dict specifying keys "host" and ' +
                                 'optionally "db_profile"') % cluster)
            else:
                hosts.append(host)
    else:
        raise Exception(('Error in ~/.oshrc: ' +
                         'osh.remote.%s.hosts must be a list or dict') % cluster)
    return (config_user, hosts)

def _has_profile(cluster):
    config = _cluster_configuration(cluster)
    return config[0] or config[1]

def _print_command_output(command, stdout, stderr):
    if stdout:
        print 'stdout from %s' % command
        for line in stdout:
            sys.stdout.write(line)
    if stderr:
        print 'stderr from %s' % command
        for line in stderr:
            sys.stdout.write(line)

def spawn(command):
    read, write, err = popen2.popen3(command)
    _print_command_output(command, read.readlines(), err.readlines())
    read.close()
    write.close()
    err.close()

def ssh(user, host, command):
    ssh_command = ('ssh %s -T -o StrictHostKeyChecking=no -l %s "%s" ' %
                   (host, user, command))
    spawn(ssh_command)

def copyup(files, user, host, remoteDir, options = ''):
    if type(files) == types.ListType or type(files) == types.TupleType:
        files = ' '.join(files)
    scp_command = 'scp %s %s %s@%s:%s' % (options, files, user, host, remoteDir)
    spawn(scp_command)

def copydown(user, host, file, local_dir, options = ''):
    scp_command = 'scp %s %s@%s:%s %s' % (options, user, host, file, local_dir)
    spawn(scp_command)

class _Host:

    _name = None
    _address = None
    _db_profile = None

    def __init__(self, x, y = None, z = None):
        self._name = x
        if y is None:
            self._address = x
        else:
            self._address = y
        if z:
            self._db_profile = z

    def __str__(self):
        if self._db_profile:
            return '%s(%s, %s)' % (self._name, self._address, self._db_profile)
        else:
            return '%s(%s)' % (self._name, self._address)

    def name(self):
        return self._name

    def address(self):
        return self._address

    def db_profile(self):
        return self._db_profile

class _RemoteOpThread(threading.Thread):

    _remoteop = None
    _host = None

    def __init__(self, remoteop, host):
        threading.Thread.__init__(self)
        self._remoteop = remoteop
        self._host = host

    def run(self):
        try:
            self._remoteop.action(self._host)
        except Exception, e:
            assert False, str(e)

    def id(self):
        return self._id

    def terminating_exception(self):
        return self._remoteop.terminating_exception()


class RemoteOp(Generator):
    """Base class of osh commands that run remotely. Such commands must
    be generators, (might want to permit piping of input to remote commands
    someday). Remote execution is done across all nodes in parallel.
    """

    _cluster = None
    _user = None
    _hosts = None
    _scp_options = None
    _terminating_exception = None

    # object interface
    
    def __init__(self, flag_spec, anon_range):
        Generator.__init__(self, flag_spec, anon_range)
        

    # BaseOp interface

    def setup(self, cluster_required = True):
        args = self.args()
        cluster = args.string_arg('-c')
        if cluster:
            # cluster could be one of the following:
            # - cluster defined in profile, e.g. foo
            # - cluster and pattern, e.g. foo:116
            # - neither -- interpret as a hostname
            cluster, pattern = (cluster.split(':') + [None])[:2]
            if _has_profile(cluster):
                user, hosts = _cluster_configuration(cluster)
                if pattern:
                    hosts = filter(lambda host: host.name().find(pattern) >= 0,
                                   hosts)
                # Allow for command line overrides
                if not user:
                    user = args.string_arg('-u')
            else:
                # Cluster is specified on command line but not configured
                # in .oshrc. Interpret the cluster as a hostname.
                hosts = [_Host(cluster)]
                user = args.string_arg('-u')
        else:
            # Cluster not specified.
            if cluster_required:
                # Use default in .oshrc.
                cluster = config_value('remote')
                if cluster:
                    user, hosts = _cluster_configuration(cluster)
                    # Allow for command line overrides
                    user = args.string_arg('-u')
            else:
                user, hosts = (None, None)
        if not user:
            user = os.environ['USER']
        # Convert hosts to list of _Host objects
        self._cluster, self._user, self._hosts, = (cluster, user, hosts)
        # Get scp options
        self._scp_options = ''
        for flag in SCP_OPTIONS:
            self._check_option(flag)
        if self._scp_options:
            self._scp_options = '-' + self._scp_options
        # Check that we have everything
        if cluster_required and (not self._cluster or
                                 not self._user or
                                 not self._hosts):
            self.usage()


    # Generator interface

    def execute(self):
        threads = []
        id = 0
        for host in self.hosts():
            thread = _RemoteOpThread(self, host)
            thread.start()
            threads.append(thread)
        for thread in threads:
            while thread.isAlive():
                thread.join(1.0)
                thread_termination = thread.terminating_exception()
                if thread_termination:
                    raise thread_termination



    # RemoteOp interface

    def action(self):
        """Method executed on each remote node.
        """
        assert False

    def terminating_exception(self):
        return self._terminating_exception
    
    def set_terminating_exception(self, terminating_exception):
        self._terminating_exception = terminating_exception

    def error(self, host, error):
        traceback.print_exc(file = sys.stderr)
        message = 'Error for host %s: %s' % (host, error)
        self.send_error(message)

    # For use by subclasses

    def user(self):
        return self._user

    def hosts(self):
        return self._hosts

    def scp_options(self):
        return self._scp_options

    # For use by this class

    def _check_option(self, option):
        if self.args().flag('-%s' % option):
            self._scp_options += option

class RemoteException(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)
